(use matchable)
(use hfs+)
(use base64)
(use files)
(use posix)
(cond-expand
 (compiling
  ;; to avoid local .so dep; yucky
  (include "posix-extras.scm"))
 (else
  (require-library posix-extras)))
(import (only posix-extras stat/ifmt stat/iflnk stat/ififo stat/ifsock))

(include "xattr-common.scm")
(import xattr-common)
(include "xattr-args.scm")
(import xattr-args)

;;; Usage

(define (usage-error . msg)
  (define (default key)
    (sprintf "[default: ~A]" (config-default key)))
  (with-output-to-port (current-error-port)
    (lambda ()
      (when (pair? msg)
        (print (apply conc "error: " msg)))
      (print "usage: " (program-name) " [options] [file-base] [xattr-base]")
      (print "
      --attr-mode=MODE       attribute input mode " (default 'attr-mode) "
                              dedicated:   .@@/._@filename
                              individual:  ._@filename
                              unified:     single file (UNIMPLEMENTED)
                              appledouble: copy Finder Info + Resource Fork
                                           from ._filename in AppleDouble format
                                           using our own packing routines
                              copyfile:    copy all HFS+ metadata from ._filename
                                           using native OS X AppleDouble packer
                                           (may not be 100% reliable on 10.4)
      --list-format=FMT      format of input file list " (default 'list-format) "
                              rsync-list-only: output of `rsync --list-only`
                              find: output of `find -print`
                              find0: output of `find -print0`
      --no-specials          skip fifos and sockets
      --no-symlinks          skip symlinks (recommended for copyfile on 10.4)
  -n, --dry-run              describe operations only
  -v, --verbose              verbose progress report
  -h, --help                 usage information
  -V, --version              version information

  Reads a list of files given on stdin and applies the corresponding
  serialized HFS+ metadata to each.  Input filenames have any leading /
  stripped and are relative to 'file-base' (i.e. we chdir to file-base
  at startup); file-base defaults to the current directory.  'xattr-base'
  is prefixed to attribute filenames, and defaults to empty; by setting
  it, one may apply attributes from a parallel directory structure.
")))
  (exit 1))
(args-error usage-error)   ; install args error handler -- yuck

(define (verbose . msg)
  (when (config 'verbose)
    (with-output-to-port (current-error-port)
      (lambda ()
        (print (apply conc msg))))))

;;; Processing

;; Exactly the same as split-xattr version.  Consider merging
(define (process-files-on-port port format process-attrs)
  (match (list-format format)
         ((_ read-record parse-record)
          (let loop ()
            ;;       (flush-output (current-output-port))     ;; for debugging
            ;;       (flush-output (current-error-port))      ;; for debugging
            (let ((record (read-record port)))
              (if (eof-object? record)
                  'done
                  (let ((filename (parse-record record)))
                    (process-attrs filename)
                    (loop))))))
         (_ (error "no such list format" format))))

;;; input modes

(define (set-extended-attributes! filename attrs)
  ;; Warning: If resource fork is modified (com.apple.ResourceFork), it will not
  ;; be truncated.  You must clear it beforehand.  Here, we forcibly
  ;; clear all existing attributes beforehand.
  (clear-extended-attributes! filename #:no-follow)
  (for-each
   (lambda (x)
     (match x
            ((key ('base64 val))
             (set-extended-attribute! filename key
                                      (base64-decode val) #:no-follow))
            ((key val)
             (set-extended-attribute! filename key
                                      val #:no-follow))))
   attrs))

(define attr-same-dir
  (lambda (filename base renamer unserializer)
    (condition-case
     (let* ((xattr-file (renamer filename base))
            (stat (file-stat filename #t))       ; missing input file -> error
            (mode (vector-ref stat 1))
            (ifmt (bitwise-and mode stat/ifmt)))
       (cond ((not (file-exists? xattr-file))
              #f) ; silently skip files without corresponding attribute records
             ((and (or (= stat/ififo ifmt)
                       (= stat/ifsock ifmt))
                   (config 'no-specials))
              (verbose "skip special " filename))
             ((and (= stat/iflnk ifmt)
                   (config 'no-symlinks))
              (verbose "skip symlink " filename))
             (else
              (unserializer xattr-file filename))))          ; no xattrs
     (e (exn file)
        (warning
         (conc "Error joining attributes: "
               ((condition-property-accessor 'exn 'message) e)
               " "
               (let ((args ((condition-property-accessor 'exn 'arguments) e)))
                 (if (null? args) "" args)))
         filename)))))

(define (xattr-unserializer:sexp xattr-file output-filename)
  (let ((attrs (read-file xattr-file))) ; failure means unreadable or bad format
    (cond ((assq 'xattr attrs)
           => (lambda (record)
                (let ((attrs (cdr record)))
                  (verbose "join xattr " output-filename)
                  (unless (config 'dry-run)
                    (set-extended-attributes! output-filename attrs))
                  #t)))
          (else #f))))

(define (xattr-unserializer:appledouble xattr-file output-filename)
  (error "appledouble unserializer unimplemented"))

(define (xattr-unserializer:copyfile xattr-file output-filename)
  (verbose "unpack xattr " output-filename)
  (unless (config 'dry-run)
    (copyfile xattr-file output-filename #:unpack #:metadata #:no-follow)))

(define attr-unified
  (lambda (filename port)
    ;; unified mode requires completely different structure: normally we use
    ;; process-files-on-port on input to obtain files, but here
    ;; we need to parse the xattr records.  probably need to split out.
    ;; one weird approach would be to have a list-format unified-list which
    ;; would have special read-record and parse-record, and rather than
    ;; passing a plain filename, pass the entire filename/xattr sexp here.
    ;; port would be unnecessary.  then, when implementing actual read of
    ;; filenames from stdin, we can call into this.
    ;; --
    ;; Cool idea, but somewhat broken; for example we'd expect this new list-format
    ;; "unified-list" to be usable with any attr-mode, but it won't work due
    ;; to the filename / xattr-record mismatch.
    ;; --
    ;; Also note, when attr-mode is unified and list-format is not, we
    ;; probably need to read the whole list in at once into a hash
    ;; table, so that we can process the large attribute file linearly
    ;; (as it is not a database).  This is once again a different paradigm.
    (usage-error "unified attribute mode unimplemented")))

(define (attr-mode mode base)
  (cond ((assq mode
               `((dedicated . ,(lambda (filename)
                                     (attr-same-dir filename base
                                                      xattr-filename:dedicated-prefix
                                                      xattr-unserializer:sexp)))
                 (inline . ,(lambda (filename)
                              (attr-same-dir filename base
                                              xattr-filename:individual-prefix
                                              xattr-unserializer:sexp)))
				 ;; Unified won't currently work because both file list and
				 ;; attributes are hardcoded to come in on stdin. [Unless
				 ;; attributes on stdin means "join every file we can find."]
                 (unified . ,(lambda (filename)
                               (when base     ;; FIXME need only be done once; same as port!
                                   (usage-error
                                    "xattr-base may not yet be specified in unified mode"))
                               (attr-unified filename
                                             (current-input-port))))
                 (appledouble . ,(lambda (filename)
                                   (attr-same-dir filename base
                                                  xattr-filename:appledouble
                                                  xattr-unserializer:appledouble)))
                 (copyfile . ,(lambda (filename)
                                (attr-same-dir filename base
                                               xattr-filename:appledouble
                                               xattr-unserializer:copyfile)))
                 ;; debugging -- just print filenames
                 (debug-print . ,(lambda (fn) (write fn) (newline)))
                 ))
         => cdr)
        (else (usage-error "no such input mode" mode))))

;;; config

;; The parsed arguments and default config are both plain alists so
;; (append args defaults) provides a config alist with args
;; overriding.  An arg value is always a string or #f and type
;; conversion is done manually; so default config values must also be
;; strings.  A missing key is an error so keys defaulting to #f must
;; still be provided in default-config; this helps catch misspelled or
;; erroneous keys at runtime.

;; Type conversion on command-line-args could and probably should
;; be done at parse-args time, but input validation should be
;; done later on the final config struct.

(define *config* '())
(define *default-config*
  `((list-format . "rsync-list-only")
    (attr-mode . "dedicated")
    (verbose . #f)
    (dry-run . #f)
    (no-specials . #f)
    (no-symlinks . #f)))
(define *options*
  (list (option '("list-format") #t)
        (option '("attr-mode") #t)
        (option '("no-symlinks"))
        (option '("no-specials"))
        (option '("verbose" #\v))
        (option '("dry-run" #\n))
        (option '("help" #\h #\?) #f
                (lambda _ (usage-error)))
        (option '("version" #\V) #f
                (lambda _
                  (fprintf (current-error-port "0.0\n"))
                  (exit 0)))
        ))

(define (configure! args opts)
  (match-let (((options operands)
               (parse-arguments args opts)))
    (set! *config*
          (append
           `((file-base . ,(if (pair? operands)
                               (car operands)
                               "."))
             (xattr-base . ,(if (and (pair? operands)
                                     (pair? (cdr operands)))
                                (cadr operands)
                                #f)))
           options
           *default-config*))))

(define (config key)
  (cond ((assq key *config*) => cdr)
        (else (usage-error "missing configuration key: " key))))
(define (config-default key)
  (cond ((assq key *default-config*) => cdr)
        (else (error "missing default configuration key" key))))

;;; main

(args-error usage-error)   ; install args error handler -- yuck
(configure! (command-line-arguments) *options*)

(when (config 'dry-run)
  (fprintf (current-error-port)
           "[** Dry run; no changes will be made **]\n"))

(current-directory (config 'file-base))
(time
 (process-files-on-port (current-input-port)
                        (string->symbol (config 'list-format))
                        (attr-mode (string->symbol (config 'attr-mode))
                                    (config 'xattr-base))))

;; Take attributes we saved from ~/tmp into ~/tmp.xattr and propagate them into tmp.2, which is a copy of tmp.
;; Could use rsync -a ~/tmp.2/ as well, does not matter

;; time /opt/local/bin/rsync -a ~/tmp/ --filter '- .@@' --filter ': .rlb-rules' | csi4 -script ./join-xattr.scm ~/tmp.2 ~/tmp.xattr
