;; TODO: Handle conflicts with existing files
;;       Attribute failure on socket/pipe should be ignored (possible?)
;;       Symlink xattrs should get perms/owner of referenced file [and handle dangling!]
;;           lrwxrwxrwx   1 jim  jim  7 Aug 13  2007 .bash_profile@ -> .bashrc
;;           -rw-rw-rw- 1 501 501 96 Aug 13  2007 .@_.bash_profile               ; wrong
;;       Unified output should go to arbitrary port (xattr-base filename, probably).
;;          Currently it always goes to stdout.

;; Attribute files are set to ctime of original file to avoid spurious recopy w/o --checksum
;;  (Resource fork update sets both mtime and ctime)
;; Permissions on attribute files are taken from original file
;; Automatically-created directories get parent perms
;; Configurable by command-line-args

;; FIXME (chicken): change-directory and file-exists? does not handle ~ (##sys#expand-home-path is broken)


(use regex)
(use matchable)
(use hfs+)
(use base64)
(use files)
(use posix)
(import irregex)
(require-library srfi-1)
(import (only srfi-1 partition filter-map))

(cond-expand
 (compiling
  ;; to avoid local .so dep; yucky
  (include "posix-extras.scm"))
 (else
  (require-library posix-extras)))
(import (only posix-extras change-file-times stat/ififo stat/ifsock stat/ifmt stat/iflnk))

(include "xattr-common.scm")
(import xattr-common)
(include "xattr-args.scm")
(import xattr-args)
(include "xattr-appledouble.scm")
(import xattr-appledouble)

(define (usage-error . msg)
  (define (default key)
    (sprintf "[default: ~A]" (config-default key)))
  (with-output-to-port (current-error-port)
    (lambda ()
      (when (pair? msg)
        (print (apply conc "error: " msg)))
      (print "usage: " (program-name) " [options] [file-base] [xattr-base]")
      (print "
      --attr-mode=FMT        attribute output mode " (default 'attr-mode) "
                              dedicated:   .@@/._@filename
                              individual:  ._@filename
                              unified:     single file (to stdout)
                              appledouble: copy Finder Info + Resource Fork
                                           to ._filename in AppleDouble format
                                           using our own packing routines
                              copyfile:    copy all HFS+ metadata to ._filename
                                           using native OS X AppleDouble packer
                                           (may not be 100% reliable on 10.4)
      --list-format=FMT      format of input file list " (default 'list-format) "
                              rsync-list-only: output of `rsync --list-only`
                              find: output of `find -print`
                              find0: output of `find -print0`
      --no-specials          skip fifos and sockets
      --no-symlinks          skip symlinks (recommended for copyfile on 10.4)
  -n, --dry-run              describe operations only
  -v, --verbose              verbose progress report
  -h, --help                 usage information
  -V, --version              version information

  Reads HFS+ metadata from a list of files given on stdin and outputs it
  to one or many files, depending on the selected output format.
  Input filenames are relative to 'file-base' (i.e. we chdir to file-base
  at startup), which defaults to the current directory.  'xattr-base'
  is prefixed to output filenames, and defaults to empty; by setting
  it, one may output to a parallel directory structure.
")))
  (exit 1))

(define (verbose . msg)
  (when (config 'verbose)
    (with-output-to-port (current-error-port)
      (lambda ()
        (print (apply conc msg))))))

;; Obtain HFS+ attributes and return in the format:
;; ((com.apple.FinderInfo . string-data
;;  (com.apple.ResourceFork . string-data))
(define (get-attribute-record file)
  (map (lambda (x)
         (cons (string->symbol x)
               (get-extended-attribute file x #:no-follow)))
       (sort (list-extended-attributes file #:no-follow) string<?)))

;;; Output formatters

;; These take an ATTRIBUTE-RECORD, a la get-attribute-record, and write it
;; to OUTPUT-PORT in a particular format.  We provide formatters for
;; the standard sexp output, and AppleDouble.

;; (xattr (com.apple.FinderInfo (base64 "base64-string")
;;        (com.apple.ResourceFork (base64 "base64-string")))
(define (xattr-format:sexp attribute-record output-port)
  (pp `(xattr ,@(map (lambda (x)
                       (match-let (((name . data) x))
                         `(,name (base64 ,(base64-encode data)))))
                     attribute-record))
      output-port))

(define (xattr-format:appledouble attribute-record output-port)
  ;; Obtain finder info and resource fork if present, reorder them
  ;; so resource fork is last, convert to AppleDouble entry
  ;; structure, encode to AppleDouble, and output.

  ;; (Brokenly, neither OS X nor FixupResourceForks
  ;; understand AppleDouble files unless they contain both a Finder
  ;; Info section and a Resource Fork, and nothing else.  Dummy
  ;; default values must be provided.  Currently, we do not write
  ;; anything if no valid attributes exist, although this will
  ;; create an invalid 0-byte file.)

  ;; So far, the only unhandled attribute I've seen is 'com.apple.TextEncoding
  ;; However, Apple tar encodes extended attributes into finder-info
  ;; and resource-fork in an unknown format.  FixupResourceForks can't
  ;; decode this correctly.

  (define finder-dummy   ;; 32 bytes of zeros
    (string-append
     "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
     "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"))

  (receive (good bad)
      (partition (lambda (x)
                   (or (eq? 'com.apple.FinderInfo (car x))
                       (eq? 'com.apple.ResourceFork (car x))))
                 attribute-record)
    (for-each (lambda (x) (warning "invalid attribute" (car x)))
              bad)
    (if (pair? good)
        (appledouble-encode
         (list (cons 'finder-info
                     (cond ((assq 'com.apple.FinderInfo good) => cdr)
                           (else finder-dummy)))
               (cons 'resource-fork
                     (cond ((assq 'com.apple.ResourceFork good) => cdr)
                           (else ""))))
         output-port))))

;;; Output

(define (process-files-on-port port format process-attrs)
  (match (list-format format)
         ((_ read-record parse-record)
          (let loop ()
            ;;       (flush-output (current-output-port))     ;; for debugging
            ;;       (flush-output (current-error-port))      ;; for debugging
            (let ((record (read-record port)))
              (if (eof-object? record)
                  'done
                  (let ((filename (parse-record record)))
                    (process-attrs filename)
                    (loop))))))
         (_ (error "no such list format" format))))

(define (output-unified filename port)
  (cond ((not filename)                 ; parse failure
         #f)
        (else
         ;; (printf ";; file: ~S\n" filename)
         (condition-case
          (let* ((stat (file-stat filename #t))
                 (mode (vector-ref stat 1))
                 (ifmt (bitwise-and mode stat/ifmt)))
            (cond
             ((and (or (= stat/ififo ifmt)
                       (= stat/ifsock ifmt))
                   (config 'no-specials))
              (verbose "skip special " filename))
             ((and (= stat/iflnk ifmt)
                   (config 'no-symlinks))
              (verbose "skip symlink " filename))
             (else
              (let ((attrs (get-attribute-record filename)))
                (when (pair? attrs)
                  (verbose "unified xattr " filename)
                  (pp `(file ,filename (xattr ,@attrs))
                      port))))))
          (e (exn)    ; FIXME exn hfs+; and record error for exit code
             (warning (conc "Error reading attributes: "
                            ((condition-property-accessor 'exn 'message) e)) filename)
             (fprintf port ";; Error: ~S ~A\n"
                      filename
                      ((condition-property-accessor 'exn 'message) e))))
         #f)))  ;; ?

(define (call-with-output-file* fn proc)
  (let ((out (open-output-file fn)))
    (let ((rv (handle-exceptions exn
                  (begin (close-output-port out)
                         (abort exn))
                (proc out))))
      (close-output-port out)
      rv)))

;; (renamer FILENAME BASE): Rename input filename to output filename based at BASE.
;; (formatter RECORD PORT): Write formatted attribute record to PORT.
(define output-same-dir
  (let ((safe-umask
         (bitwise-ior perm/irusr perm/irgrp perm/iroth
                      perm/iwusr perm/iwgrp perm/iwoth)))
    (lambda (filename base renamer formatter)
      (cond ((not filename)
             #f)
            (else
             (condition-case
              ;; If the input file disappears, an error is currently raised
              (let* ((stat (file-stat filename #t))
                     (mode (vector-ref stat 1))
                     (uid (vector-ref stat 3))
                     (gid (vector-ref stat 4))
                     (ctime (vector-ref stat 7))
                     (ifmt (bitwise-and mode stat/ifmt)))
                (cond
                 ((and (or (= stat/ififo ifmt)
                           (= stat/ifsock ifmt))
                       (config 'no-specials))
                  (verbose "skip special " filename))
                 ((and (= stat/iflnk ifmt)
                       (config 'no-symlinks))
                  (verbose "skip symlink " filename))
                 (else
                  (let ((attrs (get-attribute-record filename)))
                    (when (pair? attrs)
                      (let ((output-filename (renamer filename base)))
                        (verbose "split xattr " output-filename)
                        (unless (config 'dry-run)
                          (and-let* ((parent-dir (pathname-directory output-filename)))
                            ;; `mkdir -p` with reasonable permissions on intermediate dirs
                            (create-directory-chain parent-dir))
                          (call-with-output-file* ; FIXME: conflict resolve with existing files
                           output-filename
                           (lambda (out)
                             (formatter attrs out)))
                          ;; Error here might leave an orphaned file; consider deleting.
                          ;; Likely to occur only for regular user chowning away from himself,
                          ;; which may even be benign; consider ignoring. (!)
                          (change-file-times output-filename ctime ctime)
                          (change-file-mode output-filename (bitwise-and mode safe-umask))
                          (change-file-owner output-filename uid gid))))))))
              (e (exn file) ; FIXME record error for exit code
                 (warning (conc "Error reading attributes: "
                                ((condition-property-accessor 'exn 'message) e)
                                " "
                                (let ((args ((condition-property-accessor 'exn 'arguments) e)))
                                  (if (null? args) "" args)))
                          filename)))
             #f)))))    ; FIXME I think this is a relic

;;
;; You should use 'no-symlinks on 10.4, as the #:no-follow option to copyfile
;; is ignored for pack/unpack.
;; NB This is now similar enough to output-same-dir to be merged (as in join-xattr)
(define output-copyfile
  (let ((safe-umask
         (bitwise-ior perm/irusr perm/irgrp perm/iroth
                      perm/iwusr perm/iwgrp perm/iwoth)))
    (lambda (filename base renamer)
      (cond ((not filename)
             #f)
            (else
             (condition-case
              (let* ((stat (file-stat filename #t))
                     (mode (vector-ref stat 1))
                     (uid (vector-ref stat 3))
                     (gid (vector-ref stat 4))
                     (ctime (vector-ref stat 7))
                     (ifmt (bitwise-and mode stat/ifmt)))
                (cond ((and (or (= stat/ififo ifmt)
                                (= stat/ifsock ifmt))
                            (config 'no-specials))
                       (verbose "skip special " filename))
                      ((and (= stat/iflnk ifmt)
                            (config 'no-symlinks))
                       (verbose "skip symlink " filename))
                      (else
                       (let ((output-filename (renamer filename base)))
                         (when (> (copyfile filename #f #:check #:metadata #:no-follow)
                                  0) ;; rv > 0 indicates there is metadata to copy
                           (verbose "pack " output-filename)
                           (unless (config 'dry-run)
                             (and-let* ((parent-dir (pathname-directory output-filename)))
                               ;; `mkdir -p` with reasonable permissions on intermediate dirs
                               (create-directory-chain parent-dir))
                             (copyfile filename output-filename #:pack #:metadata #:no-follow)
                             ;; Seems like copyfile doesn't assign ctime to packed file nor
                             ;; does it set the mode correctly (might use umask).  It does
                             ;; set owner, but we do it anyway.
                             (change-file-times output-filename ctime ctime)
                             (change-file-mode output-filename (bitwise-and mode safe-umask))
                             (change-file-owner output-filename uid gid)
                             ))))))
              (e (exn file)         ; FIXME record error for exit code
                 (warning (conc "Error reading attributes: "
                                ((condition-property-accessor 'exn 'message) e)
                                " "
                                (let ((args ((condition-property-accessor 'exn 'arguments) e)))
                                  (if (null? args) "" args)))
                          filename)))
             #f)))))

;; Like (create-directory path #t) from Unit files, but sets the mode and
;; owner of intermediate directories to be that of the parent.
;; create-directory traverses the full absolute path; this will not.
(define create-directory-chain
  (let ((safe-umask (bitwise-ior perm/irusr perm/irgrp perm/iroth
                                 perm/iwusr perm/iwgrp perm/iwoth
                                 perm/ixusr perm/ixgrp perm/ixoth)))
    (lambda (path)
      (let ((oldwd (current-directory)))
        (handle-exceptions exn
            (begin (current-directory oldwd)
                   (signal exn))
          (when (absolute-pathname? path)
            (current-directory "/"))
          (let loop ((chain (string-split path "/")))
            (if (null? chain)
                #t
                (let ((dir (car chain)))
                  ;; Handling of ~ in filenames is also broken due to auto-expand
                  ;; below; use C or prepend "./" to directory name
                  (cond ((directory? dir) ;; better: mkdir and ignore existing dirs; requires C
                         ;; Existing directories not reported in verbose mode.
                         (current-directory dir)
                         (loop (cdr chain)))
                        (else
                         (let* ((stat (file-stat "."))
                                (uid (vector-ref stat 3))
                                (gid (vector-ref stat 4))
                                (mode (vector-ref stat 1))
                                (mask (bitwise-and mode safe-umask)))
                           (cond ((config 'dry-run)
                                  ;; Special : updating current-directory won't work when
                                  ;; directories do not exist.
                                  ;; Note: spurious output anyway as we try to create the
                                  ;; directory for every single file.
                                  (verbose "mkdir " uid " " gid " " (number->string mask 8)
                                           " ... "
                                           path))   ; just skip to the end
                                 (else
                                  (verbose "mkdir " uid " " gid " " (number->string mask 8) " "
                                           (make-pathname (current-directory)
                                                          dir))
                                  (create-directory dir) ; no default mode allowed!
                                  (change-file-mode dir mask)
                           ;; Again, failure to change owner is not necessarily a problem
                                  (change-file-owner dir uid gid)
                                  (current-directory dir)
                                  (loop (cdr chain)))))))))))
        (current-directory oldwd)))))

;; return proc of 1 argument FILENAME which obtains extended attributes from
;; FILENAME and writes them out to the appropriate file or port
;; For 'unified' the port must be reused for each filename without closing;
;; easiest is pass open port as 'base', which negates this abstraction. :(
(define (output-mode mode base)
  (cond ((assq mode
               `((dedicated . ,(lambda (filename)
                                 (output-same-dir filename base
                                                  xattr-filename:dedicated-prefix
                                                  xattr-format:sexp)))
                 (inline . ,(lambda (filename)
                              (output-same-dir filename base
                                               xattr-filename:individual-prefix
                                               xattr-format:sexp)))
                 (unified . ,(lambda (filename)
                               (output-unified filename
                                               (current-output-port))))
                 (appledouble . ,(lambda (filename)
                                   (output-same-dir filename base
                                                    xattr-filename:appledouble
                                                    xattr-format:appledouble)))
                 (copyfile . ,(lambda (filename)
                                (output-copyfile filename base
                                                 xattr-filename:appledouble)))
                 ))
         => cdr)
        (else (usage-error "no such output mode: " mode))))

; /opt/local/bin/rsync -a ../chicken-4 --filter ': .rlb-rules' | csi4 -script split-xattr.scm

;;; config

(define *options*
  (list (option '("list-format") #t)
        (option '("attr-mode") #t)
        (option '("no-specials"))
        (option '("no-symlinks"))
        (option '("dry-run" #\n))
        (option '("verbose" #\v))
        (option '("help" #\h #\?) #f
                (lambda _ (usage-error)))
        (option '("version" #\V) #f
                (lambda _
                  (fprintf (current-error-port) "0.0\n")
                  (exit 0)))))
(define *default-config*
  `((list-format . "rsync-list-only")
    (attr-mode . "dedicated")
    (verbose . #f)
    (dry-run . #f)
    (no-specials . #f)
    (no-symlinks . #f)))
(define *config* '())

(define (configure! args opts)
  (match-let (((options operands)
               (parse-arguments args opts)))
    (set! *config*
          (append
           `((file-base . ,(if (pair? operands)
                               (car operands)
                               "."))
             (xattr-base . ,(if (and (pair? operands)
                                     (pair? (cdr operands)))
                                (cadr operands)
                                #f)))
           options
           *default-config*))))

(define (config key)
  (cond ((assq key *config*) => cdr)
        (else (usage-error "missing configuration key: " key))))
;; (define (config-option key)
;;   (cond ((assq key *config*) => cdr)
;;         (else #f)))
(define (config-default key)
  (cond ((assq key *default-config*) => cdr)
        (else (error "missing default configuration key" key))))
(define (unimplemented option name arg . _)
  (usage-error "unimplemented option: " name))

;;; main

(args-error usage-error)   ; install args error handler -- yuck
(configure! (command-line-arguments) *options*)

(when (config 'dry-run)
  (fprintf (current-error-port)
           "[** Dry run; no changes will be made **]\n"))

(current-directory (config 'file-base))
(time
 (process-files-on-port (current-input-port)
                        (string->symbol
                         (config 'list-format))
                        (output-mode (string->symbol
                                      (config 'attr-mode))
                                     (config 'xattr-base))))

;; /opt/local/bin/rsync -a ~ --filter ': .rlb-rules' | csi4 -script split-xattr.scm /Users >| ~/tmp/hfs.out1 2>|~/tmp/hfs.err1

;; # time /opt/local/bin/rsync -a / --exclude-from=/usr/local/etc/rlbackup.exclude --filter ': .rlb-rules' | /Users/jim/local/bin/csi4 -script ./split-xattr.scm / > /tmp/hfs-backup1.out 2>/tmp/hfs-backup1.err
;; real    0m48.922s user    0m33.859s sys     0m18.426s


;; /opt/local/bin/rsync -a --list-only --filter='dir-merge .rlb-rules' --filter '- .@@' --exclude-from=/usr/local/etc/rlbackup.exclude --relative / | /Users/jim/local/bin/csi4 -I /Users/jim/scheme/xattr -script /Users/jim/scheme/xattr/split-xattr.scm / 2> /tmp/hfs-backup2.err


;; gross -- obtain only .@@/* attribute files (and all other directories, regardless of whether they are interceding)
;; /opt/local/bin/rsync -a --list-only --filter='dir-merge .rlb-rules' --exclude-from=/usr/local/etc/rlbackup.exclude --filter "+ **/.@@" --filter "+ */" --filter "+ .@@/*" --filter "- *" --relative /


;; /opt/local/bin/rsync -av -c -x --numeric-ids -H --filter='dir-merge .rlb-rules' --exclude-from=/usr/local/etc/rlbackup.exclude --filter "+ **/.@@" --filter "+ */" --filter "+ .@@/*" --filter "- *" --relative / rlb_amaranth@alice:snapshot/0
