;;; AppleDouble encoding and decoding
;; Copyright (c) 2009 Jim Ursetto.  All rights reserved.
;; License: BSD.

(module xattr-appledouble (appledouble-encode
                           appledouble-decode
                           symbol->appledouble-entry-id
                           appledouble-entry-id->symbol)

(import scheme chicken)
(import (only extras write-string))
(import (only ports with-output-to-port))
(use matchable)

;;; Header constants

(define-constant +magic-number+ #x00051607)
(define-constant +appledouble-version+ #x00020000)
(define-constant +filler+ "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00")

;;; Simple binary I/O

(define (char x)
  (integer->char (bitwise-and x 255)))
(define (write-u32 x)
  (write-char (char (arithmetic-shift x -24)))
  (write-char (char (arithmetic-shift x -16)))
  (write-char (char (arithmetic-shift x  -8)))
  (write-char (char x)))
(define (write-u16 x)
  (write-char (char (arithmetic-shift x  -8)))
  (write-char (char x)))
(define (write-u8 x)
  (write-char (char x)))

;;; Entry IDs

;; Convert AppleDouble entry ID integer INT to a human-readable
;; symbol.  If no corresponding symbol is known, returns INT.
(define (appledouble-entry-id->symbol int)
  (case int
    ((1) 'data-fork)
    ((2) 'resource-fork)
    ((3) 'real-name)
    ((4) 'comment)
    ((5) 'icon/black-and-white)
    ((6) 'icon/color)
    ((8) 'file-dates-info)
    ((9) 'finder-info)
    ((10) 'macintosh-file-info)
    ((11) 'prodos-file-info)
    ((12) 'msdos-file-info)
    ((13) 'short-name)
    ((14) 'afp-file-info)
    ((15) 'directory-id)
    (else int)))

;; Convert human-readable AppleDouble entry symbol SYM into an
;; integer entry ID.  If there is no corresponding entry ID,
;; an error is raised.
(define (symbol->appledouble-entry-id sym)
  (case sym
    ((data-fork) 1)
    ((resource-fork) 2)
    ((real-name) 3)
    ((comment) 4)
    ((icon/black-and-white) 5)
    ((icon/color) 6)
    ((file-dates-info) 8)
    ((finder-info) 9)
    ((macintosh-file-info) 10)
    ((prodos-file-info) 11)
    ((msdos-file-info) 12)
    ((short-name) 13)
    ((afp-file-info) 14)
    ((directory-id) 15)
    (else
     (error "no corresponding entry ID for type" sym))))

;; Encode entries in ENTRIES into AppleDouble v2 format
;; to PORT.  Returns PORT.  If PORT is not provided, a new
;; string will be created (and returned).
;; ENTRIES := '(ENTRY ENTRY ENTRY ...)
;; ENTRY   := '(ID . DATA)
;; ID      := entry ID as an integer or symbol (implicit symbol->entry-id)
;; DATA    := entry data as a string
;; This is fairly low level; no bounds checking is done on
;; numeric entry IDs, nor validity checks on entry data.

(define (appledouble-encode entries #!optional (port #f))
  (cond
   ((not port) (get-output-string
                (appledouble-encode entries (open-output-string))))
   (else
    (with-output-to-port port
      (lambda ()
        (let ((num-entries (length entries)))
          (write-u32 +magic-number+)          ;; 4
          (write-u32 +appledouble-version+)   ;; 4
          (write-string +filler+)             ;; 16
          (write-u16 num-entries)             ;; 2
          (let write-descs ((offset (+ 4 4 16 2
                                       (* num-entries (+ 4 4 4))))
                            (entries entries))
            (if (pair? entries)
                (match-let (((id . data) (car entries)))
                  (let ((len (string-length data)))
                    (write-u32 (if (symbol? id)
                                   (symbol->appledouble-entry-id id)
                                   id))
                    (write-u32 offset)
                    (write-u32 len)
                    (write-descs (+ offset len)
                                 (cdr entries)))))))
        (for-each (lambda (e)
                    (match-let (((id . data) e))
                      (write-string data)))
                  entries)
        port)))))

(define (appledouble-decode port)
  (error "appledouble decode not implemented"))

)
